/****************************************************************************
*       main.c
*
* **********************************************************************
* *     Copyright (c) by Barak Pearlmutter and Kevin Lang, 1987-99.    *
* *     Copyright (c) by Alex Stuebinger, 1998-99.                     *
* *     Placed under the terms of the GNU General Public License       *
* *     as published by the Free Software Foundation.                  *
* **********************************************************************
*
*/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include "config.h"
#include "data.h"
#include "cmdline.h"
#include "weak.h"
#include "stacks.h"
#include "worldio.h"
#include "loop.h"



int
get_byte_gender (void)
{
  /* Byte Gender Detection Routine */
  unsigned long a = 0x04030201ul;
  unsigned char *cp = (unsigned char *) &a;

  if (cp[0] == 0x01)
    return little_endian;
  else
    return big_endian;
}


#include "xmalloc.h"
unsigned char *
oak_c_string (ref_t * oakstr, int len)
{
  /* Converts an Oaklisp string, given by a pointer to its
     start and a length, to an equivalent C-string.
     The storage allocated by this routine must be free()-ed.
   */
  int i = 0;
  unsigned char *const cstring = xmalloc (len + 1);
  unsigned long temp;
  unsigned char *const cp = (unsigned char *) &temp;

  while (i + 2 < len)
    {
      temp = (unsigned long) *oakstr >> 2;
      cstring[i + 0] = cp[0];
      cstring[i + 1] = cp[1];
      cstring[i + 2] = cp[2];
      oakstr++;
      i += 3;
    }
  if (i + 1 < len)
    {
      temp = (unsigned long) *oakstr >> 2;
      cstring[i + 0] = cp[0];
      cstring[i + 1] = cp[1];
      oakstr++;
      i += 2;
    }
  else if (i < len)
    {
      temp = (unsigned long) *oakstr >> 2;
      cstring[i + 0] = cp[0];
      /* oakstr++; */
      i++;
    }
  cstring[i + 0] = '\0';
  return cstring;
}


int
main (int argc, char **argv)
{

  byte_gender = get_byte_gender ();

  parse_cmd_line (argc, argv);

  init_weakpointer_tables ();

  init_stacks ();

  read_world (world_file_name);

  new_space.size = e_next_newspace_size = original_newspace_size;
  alloc_space (&new_space, new_space.size);
  free_point = new_space.start;


  /* Set the registers to the boot code */

  e_current_method = e_boot_code;
  e_env = REF_TO_PTR (REF_SLOT (e_current_method, METHOD_ENV_OFF));
  e_code_segment = REF_SLOT (e_current_method, METHOD_CODE_OFF);
  e_pc = CODE_SEG_FIRST_INSTR (e_code_segment);

  /* Put a reasonable thing in e_bp to avoid confusing GC */
  e_bp = e_env;

  /* Tell the boot function the truth */
  e_nargs = 0;

  /* Big virtual machine interpreter loop */
  loop();

  return 0;
}
